/********************************************************************************
 * Copyright © 2014-2016 by Lorenzo Porta (Vindex17) <vindex17@outlook.it>      *
 *                                                                              *
 * This program is free software; you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as               *
 * published by the Free Software Foundation; either version 2 of               *
 * the License or (at your option) version 3 or any later version               *
 * accepted by the membership of KDE e.V. (or its successor approved            *
 * by the membership of KDE e.V.), which shall act as a proxy                   *
 * defined in Section 14 of version 3 of the license.                           *
 *                                                                              *
 * This program is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
 * GNU General Public License for more details.                                 *
 *                                                                              *
 * You should have received a copy of the GNU General Public License            *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                              *
 ********************************************************************************/


#include "taskprogressdialog.h"
#include "kfoldersync.h"
#include "kfoldersync.moc"
#include "previewdialog.h"
#include "profiledialog.h"

#include <KActionCollection>
#include <KDBusService>
#include <KIO/CopyJob>
#include <KIO/DeleteJob>
#include <KStandardAction>
#include <KStandardGuiItem>
#include <KStartupInfo>
#include <KWindowSystem>
#include <QApplication>
#include <QDir>
#include <QEventLoop>
#include <QInputDialog>
#include <QMenuBar>
#include <QProgressDialog>
#include <QPushButton>
#include <QStandardPaths>
#include <QVBoxLayout>
#include <QtConcurrent>

#define SETUP_ACTION(NAME, ICON, TEXT, RECEIVER, SLOT_FUNCTION) \
    QAction* NAME##_action = new QAction(ICON, TEXT, kac); \
    kac->addAction(QStringLiteral(#NAME), NAME##_action); \
    connect(NAME##_action, &QAction::triggered, RECEIVER, SLOT_FUNCTION, Qt::DirectConnection)

#define SETUP_CHECKBOX(NAME, ICON, TOOLTIP) \
    NAME.setIcon(ICON); \
    NAME.setToolTip(TOOLTIP); \
    additional_option_selector_layout->addWidget(&NAME)

#define SETUP_RADIOBUTTON(OPERATION_TYPE, TEXT, ICON_NAME, TOOLTIP) \
    QRadioButton& OPERATION_TYPE##_radiobutton = const_cast<QRadioButton&>(radiobutton_array.at(static_cast<std::size_t>(OperationType::OPERATION_TYPE))); \
    OPERATION_TYPE##_radiobutton.setText(TEXT); \
    OPERATION_TYPE##_radiobutton.setIcon(QIcon::fromTheme(QStringLiteral(#ICON_NAME))); \
    OPERATION_TYPE##_radiobutton.setToolTip(TOOLTIP); \
    operation_type_selector_layout->addWidget(&OPERATION_TYPE##_radiobutton); \
    connect(&OPERATION_TYPE##_radiobutton, &QRadioButton::clicked, [&]{selected_operation_type = OperationType::OPERATION_TYPE;})


KFolderSync::KFolderSync()
{
    // Connect to dbus
    static const KDBusService dbus_service(KDBusService::Unique);
    connect(&dbus_service, &KDBusService::activateRequested, this, [&]{KStartupInfo::setNewStartupId(this, KStartupInfo::startupId()); KWindowSystem::forceActiveWindow(winId());}, Qt::DirectConnection);

    // Setup widgets
    connect(&url_requester, &UrlRequester::new_urls, [&](const std::pair<QUrl,QUrl>& urls){operation_url_lists.first.append(urls.first); operation_url_lists.second.append(urls.second);});
    connect(&url_requester, &UrlRequester::deleted_row, [&](const int& i){operation_url_lists.first.removeAt(i); operation_url_lists.second.removeAt(i);});
    // ↓
    QGroupBox* const operation_type_selector = new QGroupBox(i18nc("@title:group", "Operation type"));
    QVBoxLayout* const operation_type_selector_layout = new QVBoxLayout(operation_type_selector);
    SETUP_RADIOBUTTON(bilateral_differential_backup, i18nc("@option:radio", "Bilateral differential backup"), kfoldersync-bs,
                      i18nc("@info:tooltip", "Source directories and their respective destinations will supplement each other with missing files.\nIf conflicts are detected, old files will be overwritten."));
    SETUP_RADIOBUTTON(bilateral_incremental_backup, i18nc("@option:radio", "Bilateral incremental backup"), kfoldersync-bib,
                      i18nc("@info:tooltip", "Source directories and their respective destinations will supplement each other with missing files.\nIf conflicts are detected, old files will be renamed."));
    SETUP_RADIOBUTTON(differential_backup, i18nc("@option:radio", "Differential backup"), kfoldersync-db,
                      i18nc("@info:tooltip", "Destination directories will be updated with the content of their respective sources. Old files will be overwritten."));
    SETUP_RADIOBUTTON(full_backup, i18nc("@option:radio", "Full backup"), kfoldersync-fb,
                      i18nc("@info:tooltip", "Source directories will be copied into their respective destinations."));
    SETUP_RADIOBUTTON(incremental_backup, i18nc("@option:radio", "Incremental backup"), kfoldersync-ib,
                      i18nc("@info:tooltip", "Destination directories will be updated with the content of their respective sources. Old files will be renamed."));
    SETUP_RADIOBUTTON(synchronization, i18nc("@option:radio", "Synchronization"), kfoldersync-s,
                      i18nc("@info:tooltip", "Destination directories will be made identical to their respective sources."));
    synchronization_radiobutton.click();
    Q_ASSERT(selected_operation_type == OperationType::synchronization);
    // ↓
    QGroupBox* const additional_option_selector = new QGroupBox(i18nc("@title:group", "Additional options"));
    QVBoxLayout* const additional_option_selector_layout = new QVBoxLayout(additional_option_selector);
    SETUP_CHECKBOX(include_hidden_files_checkbox, QIcon::fromTheme(QStringLiteral("edit-select-all")), i18nc("@info:tooltip", "Hidden files and directories will be taken into account."));
    include_hidden_files_checkbox.setChecked(true);
    SETUP_CHECKBOX(log_activity_checkbox, QIcon::fromTheme(QStringLiteral("text-x-log")), i18nc("@info:tooltip", "You'll find the log file in your home directory."));
    connect(&log_activity_checkbox, &QCheckBox::toggled, this, &KFolderSync::enable_logger, Qt::DirectConnection);
    SETUP_CHECKBOX(preview_mode_checkbox, QIcon::fromTheme(QStringLiteral("fileview-preview")), i18nc("@info:tooltip", "Changes will be displayed only."));
    preview_mode_checkbox.setChecked(true);
    SETUP_CHECKBOX(quit_when_finished_checkbox, KStandardGuiItem::quit().icon(), i18nc("@info:tooltip", "%1 will quit as soon as all operations have been completed successfully.", qApp->applicationDisplayName()));
    SETUP_CHECKBOX(use_trash_checkbox, QIcon::fromTheme(QStringLiteral("user-trash")), i18nc("@info:tooltip", "Files and directories will be moved to trash instead of being deleted."));
    // ↓
    QWidget* const central_widget = new QWidget;
    QGridLayout* const central_widget_layout = new QGridLayout(central_widget);
    central_widget_layout->addWidget(&url_requester,             0, 0, 1, 3);
    central_widget_layout->addWidget(operation_type_selector,    1, 0, 1, 1);
    central_widget_layout->addWidget(additional_option_selector, 1, 1, 1, 1);
    central_widget_layout->addWidget(&exclusion_widget,          1, 2, 1, 1);

    // Interconnect widgets
    connect(&full_backup_radiobutton, &QRadioButton::toggled, &exclusion_widget, &ExclusionWidget::setDisabled, Qt::DirectConnection);
    connect(&full_backup_radiobutton, &QRadioButton::toggled, &include_hidden_files_checkbox, &QCheckBox::setDisabled, Qt::DirectConnection);
    connect(&synchronization_radiobutton, &QRadioButton::toggled, &use_trash_checkbox, &QCheckBox::setEnabled, Qt::DirectConnection);

    // Setup actions
    KActionCollection* const kac = actionCollection();
    KStandardAction::quit(qApp, SLOT(closeAllWindows()), kac);
    KStandardAction::showMenubar(menuBar(), SLOT(setVisible(bool)), kac);
    const auto& add_ksgi = KStandardGuiItem::add();
    SETUP_ACTION(add, add_ksgi.icon(), add_ksgi.text(), &url_requester, &UrlRequester::add);
    SETUP_ACTION(manage_profiles, QIcon::fromTheme(QStringLiteral("kt-queue-manager")), i18nc("@action:intoolbar", "Manage profiles"), this, &KFolderSync::manage_profiles);
    const auto& remove_ksgi = KStandardGuiItem::remove();
    SETUP_ACTION(remove, remove_ksgi.icon(), remove_ksgi.text(), &url_requester, &UrlRequester::remove);
    const auto& reset_ksgi = KStandardGuiItem::reset();
    SETUP_ACTION(reset, reset_ksgi.icon(), reset_ksgi.text(), this, &KFolderSync::reset);
    SETUP_ACTION(run, QIcon::fromTheme(QStringLiteral("system-run")), i18nc("@action:intoolbar", "Run"), this, &KFolderSync::run);

    // Setup main window
    setCentralWidget(central_widget);
    setupGUI(ToolBar | Keys | Save | Create, QStringLiteral("kfoldersyncui.rc"));
    Q_ASSERT(minimumWidth() < 1280);
    Q_ASSERT(minimumHeight() < 768);
    show();
}


void KFolderSync::enable_logger(const bool& b)
{
    if (b) {
        Q_ASSERT(logger.isNull());
        logger = new Logger(this);
        const QString& file_name = QDir::homePath() % QStringLiteral("/(%1) - ").arg(qApp->applicationDisplayName()) % QDateTime::currentDateTime().toString(Qt::SystemLocaleLongDate) % QStringLiteral(".log");
        if (Q_UNLIKELY(logger->open_file(file_name) == false)) {
            qobject_cast<QCheckBox*>(sender())->setChecked(false);
            KMessageBox::sorry(this, xi18nc("@info", "I can't create <filename>%1</filename>...<nl/>Please check your permissions.", file_name));
        }
    } else delete logger;
}


void KFolderSync::manage_profiles()
{
    const struct {
        // never change their value or existing profiles will stop working!
        typedef const char* ProfileKey;
        ProfileKey destination_urls = "D";
        ProfileKey include_hidden_files = "IHF";
        ProfileKey log_activity = "LA";
        ProfileKey operation_type = "OT";;
        ProfileKey preview_mode = "PM";
        ProfileKey pcre_pattern_list = "PCREPL";
        ProfileKey quit_when_finished = "QWF";
        ProfileKey source_urls = "S";
        ProfileKey urls_to_exclude_list = "UEL";
        ProfileKey use_trash = "UT";
    } keys;

    // Ask
    ProfileDialog profile_dialog(this);
    const auto& config = KSharedConfig::openConfig(QStringLiteral("profiles"), KConfig::SimpleConfig, QStandardPaths::AppConfigLocation);
    const auto& config_group_list = config->groupList();
    bool ok = false;
    if (Q_UNLIKELY(profile_dialog.exec() == QDialog::Rejected))
        return;

    // Exec
    QString config_group_name;
    switch (profile_dialog.selected_action) {
    case ProfileDialog::Load:
    {
        if (Q_UNLIKELY(config_group_list.isEmpty()))
            break;
        config_group_name = QInputDialog::getItem(this, QString(), i18nc("@label", "Select the profile you want to load:"), config_group_list, 0, false, &ok, Qt::Dialog);
        if (Q_LIKELY(ok)) {
            KConfigGroup config_group(config, config_group_name);
            // checkboxes
            include_hidden_files_checkbox.setChecked(config_group.readEntry(keys.include_hidden_files, include_hidden_files_checkbox.isChecked()));
            log_activity_checkbox.setChecked(config_group.readEntry(keys.log_activity, log_activity_checkbox.isChecked()));
            preview_mode_checkbox.setChecked(config_group.readEntry(keys.preview_mode, preview_mode_checkbox.isChecked()));
            quit_when_finished_checkbox.setChecked(config_group.readEntry(keys.quit_when_finished, quit_when_finished_checkbox.isChecked()));
            use_trash_checkbox.setChecked(config_group.readEntry(keys.use_trash, use_trash_checkbox.isChecked()));
            // operation type
            selected_operation_type = static_cast<OperationType>(config_group.readEntry(keys.operation_type, static_cast<int>(selected_operation_type)));
            // input url lists
            operation_url_lists.first = config_group.readEntry(keys.source_urls, decltype(operation_url_lists.first)());
            operation_url_lists.second = config_group.readEntry(keys.destination_urls, decltype(operation_url_lists.second)());
            url_requester.restore(operation_url_lists);
            // urls to exclude and pcres
            exclusion_widget.restore(config_group.readEntry(keys.pcre_pattern_list, QList<QString>()), config_group.readEntry(keys.urls_to_exclude_list, QList<QUrl>()));
        }
    } return;
    case ProfileDialog::Save:
    {
        forever {
            config_group_name = QInputDialog::getText(this, QString(), KStandardGuiItem::saveAs().text(), QLineEdit::Normal, i18nc("@item", "Profile %1", config_group_list.size()+1), &ok, Qt::Dialog);
            if (Q_UNLIKELY(ok == false))
                return;
            if (Q_UNLIKELY(config_group_list.contains(config_group_name))) {
                if (KMessageBox::questionYesNo(this, xi18nc("@info", "<resource>%1</resource> already exists. Do you want to overwrite it?", config_group_name)) == KMessageBox::Yes) {
                    config->deleteGroup(config_group_name);
                    break;
                }
            } else break;
        }
        KConfigGroup config_group(config, config_group_name);
        // checkboxes
        config_group.writeEntry(keys.include_hidden_files, include_hidden_files_checkbox.isChecked());
        config_group.writeEntry(keys.log_activity, log_activity_checkbox.isChecked());
        config_group.writeEntry(keys.preview_mode, preview_mode_checkbox.isChecked());
        config_group.writeEntry(keys.quit_when_finished, quit_when_finished_checkbox.isChecked());
        config_group.writeEntry(keys.use_trash, use_trash_checkbox.isChecked());
        // operation type
        config_group.writeEntry(keys.operation_type, static_cast<int>(selected_operation_type));
        // url lists
        config_group.writeEntry(keys.source_urls, operation_url_lists.first);
        config_group.writeEntry(keys.destination_urls, operation_url_lists.second);
        // urls to exclude and pcres
        config_group.writeEntry(keys.pcre_pattern_list, [this]{QList<QString> list; for (const auto& pcre : exclusion_widget.pcre_list()) list.append(pcre.pattern()); return list;}());
        config_group.writeEntry(keys.urls_to_exclude_list, exclusion_widget.urls_to_exclude_list());

    } return;
    case ProfileDialog::Delete:
    {
        if (Q_UNLIKELY(config_group_list.isEmpty()))
            break;
        const auto& config_group_name = QInputDialog::getItem(this, QString(), i18nc("@label", "Select the profile you want to delete:"), config_group_list, 0, false, &ok, Qt::Dialog);
        if (Q_LIKELY(ok))
            config->deleteGroup(config_group_name);
    } return;
    case ProfileDialog::None:
        Preparer::show_bug_and_exit(__FILE__, __LINE__);
    }

    // Error (this is here to avoid redundancy)
    KMessageBox::sorry(this, i18nc("@label", "There are no saved profiles..."));
}


void KFolderSync::reset()
{
    if (KMessageBox::warningContinueCancel(this, xi18nc("@info see the reset button in the toolbar to translate the verb reset accordingly",
                                                        "<warning><application>%1</application> is going to be reset.</warning>", qApp->applicationDisplayName())) == KMessageBox::Continue) {
        hide();
        const auto& config = KSharedConfig::openConfig();
        config->markAsClean();
        KIO::file_delete(QUrl::fromLocalFile(localXMLFile()))->exec();
        QFile::remove(QStandardPaths::locate(config->locationType(), config->name()));
        config->reparseConfiguration();
        new KFolderSync;
        deleteLater();
    }
}


void KFolderSync::run()
{
    // Check and setup
    const auto& number_of_operations = operation_url_lists.first.size();
    Q_ASSERT(number_of_operations == operation_url_lists.second.size());
    if (Q_UNLIKELY(number_of_operations < 1)) {
        KMessageBox::sorry(this, i18nc("@info", "You still haven't added the directories to work with..."));
        return;
    }
    TaskProgressDialog detailed_progress_dialog(this);
    if (logger)
        logger->log_message(i18nc("@info:progress", "*** Starting ***\nOperation type: %1.\nNumber of operations to carry out: %2.",
                                  radiobutton_array.at(static_cast<std::size_t>(selected_operation_type)).text().remove(QLatin1Char('&')).toLower(), number_of_operations));

    // Perform operations sequentially
    for (std::decay<decltype(number_of_operations)>::type i = 0; i < number_of_operations; ++i) {
        const auto& current_operation_source_url = operation_url_lists.first.at(i);
        const auto& current_operation_destination_url = operation_url_lists.second.at(i);
        const QScopedPointer<Preparer> preparer(new Preparer);
        const auto& temporary_connection = connect(&detailed_progress_dialog, &TaskProgressDialog::rejected, [&]{preparer->abort_preparation = true;});
        const auto& future = QtConcurrent::run(preparer.data(),
                                               &Preparer::parse,
                                               selected_operation_type,
                                               std::make_pair(current_operation_source_url, current_operation_destination_url),
                                               include_hidden_files_checkbox.isChecked(),
                                               exclusion_widget.pcre_list(),
                                               [&](){
            QStringList path_list;
            for (const QUrl& url : exclusion_widget.urls_to_exclude_list()) {
                if (current_operation_source_url.isParentOf(url))
                    path_list.append(url.path().remove(current_operation_source_url.path()));
                if (current_operation_destination_url.isParentOf(url))
                    path_list.append(url.path().remove(current_operation_destination_url.path()));
            }
            return path_list;
        }());
        detailed_progress_dialog.wait(i18nc("title:window", "Operation %1/%2", i+1, number_of_operations));
        if (logger)
            logger->log_message(i18nc("@info:progress", "Preparing operation #%1\nSource directory: %2\nDestination directory: %3", i+1,
                                      current_operation_source_url.toDisplayString(URL_DISPLAY_FORMAT), current_operation_destination_url.toDisplayString(URL_DISPLAY_FORMAT)));
        QEventLoop event_loop(this);
        while (future.isFinished() == false) {
            event_loop.processEvents(QEventLoop::WaitForMoreEvents);
            if (preparer->abort_preparation) {
                if (logger)
                    logger->log_message(KIO::buildErrorString(KIO::ERR_ABORTED, QString()));
                Q_ASSERT(detailed_progress_dialog.isHidden());
                QProgressDialog aborting_dialog(i18nc("@label", "Aborting... please wait."), QString(), 0, 0, this, Qt::Dialog);
                aborting_dialog.setCancelButton(new QPushButton(QIcon::fromTheme(QStringLiteral("security-low")), i18nc("@action:button imperative verb", "Force")));
                aborting_dialog.open();
                while (future.isFinished() == false && aborting_dialog.wasCanceled() == false)
                    event_loop.processEvents(QEventLoop::WaitForMoreEvents);
                return;
            }
        }
        const auto& preparer_results = future.result();
        if (Q_LIKELY(preparer_results.error_message.isEmpty()))
            disconnect(temporary_connection);
        else {
            detailed_progress_dialog.hide();
            KMessageBox::error(this, preparer_results.error_message);
            if (logger)
                logger->log_message(preparer_results.error_message);
            return;
        }
        if (preview_mode_checkbox.isChecked()) {
            const auto& title = detailed_progress_dialog.windowTitle();
            if (logger)
                logger->log_message(i18nc("@info:progress", "Previewing %1...", title.toLower()));
            PreviewDialog preview_dialog(this, title, preparer_results.task_deque);
            detailed_progress_dialog.hide();
            if (Q_LIKELY(preview_dialog.exec() == QDialog::Accepted)) {
                detailed_progress_dialog.open();
                continue;
            } else return;
        }
        detailed_progress_dialog.refresh(preparer_results.task_deque.size());
        QScopedPointer<KIO::Job> current_kiojob;
        connect(&detailed_progress_dialog, &TaskProgressDialog::rejected, &event_loop, [&]{current_kiojob->kill(KIO::Job::EmitResult);}, Qt::DirectConnection);
        for (const auto& current_task : preparer_results.task_deque) {
            Q_ASSERT(std::get<1>(current_task).isValid());
            switch (std::get<0>(current_task)) {
            case TaskType::Copy:
                Q_ASSERT(std::get<2>(current_task).isValid());
                current_kiojob.reset(KIO::copyAs(std::get<1>(current_task), std::get<2>(current_task), KIO::HideProgressInfo));
                if (logger)
                    logger->log_message(xi18nc("@info:progress", "Copying...<nl/>source: <filename>%1</filename><nl/>destination: <filename>%2</filename>",
                                               std::get<1>(current_task).toDisplayString(URL_DISPLAY_FORMAT), std::get<2>(current_task).toDisplayString(URL_DISPLAY_FORMAT)));
                break;
            case TaskType::Delete:
                Q_ASSERT(std::get<2>(current_task).isEmpty());
                if (use_trash_checkbox.isChecked()) {
                    current_kiojob.reset(KIO::trash(std::get<1>(current_task), KIO::HideProgressInfo));
                    if (logger)
                        logger->log_message(xi18nc("@info:progress", "Trashing <filename>%1</filename>...", std::get<1>(current_task).toDisplayString(URL_DISPLAY_FORMAT)));
                } else {
                    current_kiojob.reset(KIO::del(std::get<1>(current_task), KIO::HideProgressInfo));
                    if (logger)
                        logger->log_message(xi18nc("@info:progress", "Deleting <filename>%1</filename>...", std::get<1>(current_task).toDisplayString(URL_DISPLAY_FORMAT)));
                } break;
            case TaskType::MakeDirectory:
                Q_ASSERT(std::get<2>(current_task).isEmpty());
                current_kiojob.reset(KIO::mkdir(std::get<1>(current_task)));
                if (logger)
                    logger->log_message(xi18nc("@info:progress", "Making directory...<nl/>target: <filename>%1</filename>", std::get<1>(current_task).toDisplayString(URL_DISPLAY_FORMAT)));
                break;
            case TaskType::Rename:
                Q_ASSERT(std::get<2>(current_task).isValid());
                current_kiojob.reset(KIO::rename(std::get<1>(current_task), std::get<2>(current_task), KIO::HideProgressInfo));
                if (logger)
                    logger->log_message(xi18nc("@info:progress", "Renaming...<nl/>source: <filename>%1</filename><nl/>destination: <filename>%2</filename>",
                                               std::get<1>(current_task).toDisplayString(URL_DISPLAY_FORMAT), std::get<2>(current_task).toDisplayString(URL_DISPLAY_FORMAT)));
                break;
            case TaskType::Overwrite:
                Q_ASSERT(std::get<2>(current_task).isValid());
                current_kiojob.reset(KIO::copyAs(std::get<1>(current_task), std::get<2>(current_task), KIO::HideProgressInfo | KIO::Overwrite));
                if (logger)
                    logger->log_message(xi18nc("@info:progress", "Overwriting...<nl/>source: <filename>%1</filename><nl/>destination: <filename>%2</filename>",
                                               std::get<1>(current_task).toDisplayString(URL_DISPLAY_FORMAT), std::get<2>(current_task).toDisplayString(URL_DISPLAY_FORMAT)));
                break;
            }
            current_kiojob->setAutoDelete(false);
            current_kiojob->setUiDelegate(nullptr);
            current_kiojob->setUiDelegateExtension(nullptr);
            if (logger) {
                connect(current_kiojob.data(), &KIO::Job::infoMessage, logger, &Logger::log_kiojob_info, Qt::DirectConnection);
                connect(current_kiojob.data(), &KIO::Job::warning, logger, &Logger::log_kiojob_info, Qt::DirectConnection);
            }
            detailed_progress_dialog.register_new_kiojob(current_kiojob.data());
            connect(current_kiojob.data(), &KIO::Job::result, [&]{event_loop.exit(current_kiojob->error());});
            if (Q_UNLIKELY(event_loop.exec() != KIO::Job::NoError)) {
                detailed_progress_dialog.hide();
                auto error_msg = current_kiojob->errorString();
                if (Q_UNLIKELY(error_msg.isEmpty()))
                    error_msg = current_kiojob->detailedErrorStrings().first();
                KMessageBox::error(this, error_msg);
                if (logger)
                    logger->log_message(error_msg);
                return;
            }
        }
    }

    // Finish
    const auto& msg = i18nc("@info:progress", "*** Mission accomplished! ***");
    if (logger)
        logger->log_message(msg);
    if (quit_when_finished_checkbox.isChecked())
        qApp->exit(EXIT_SUCCESS);
    detailed_progress_dialog.hide();
    KMessageBox::information(this, msg);
}
