/********************************************************************************
 * Copyright © 2014-2016 by Lorenzo Porta (Vindex17) <vindex17@outlook.it>      *
 *                                                                              *
 * This program is free software; you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as               *
 * published by the Free Software Foundation; either version 2 of               *
 * the License or (at your option) version 3 or any later version               *
 * accepted by the membership of KDE e.V. (or its successor approved            *
 * by the membership of KDE e.V.), which shall act as a proxy                   *
 * defined in Section 14 of version 3 of the license.                           *
 *                                                                              *
 * This program is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
 * GNU General Public License for more details.                                 *
 *                                                                              *
 * You should have received a copy of the GNU General Public License            *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                              *
 ********************************************************************************/


#include "preparer.h"
#pragma GCC poison Dialog MessageBox Widget // extra checks for thread-safety
#include "preparer.moc"

#include <KIO/ListJob>
#include <KLocalizedString>
#include <QDateTime>
#include <QRegularExpressionMatch>


#define APPLY_FILTERS \
    if (udsentry_path == dot || udsentry_path == dotdot) continue; \
    bool udsentry_must_be_excluded = false; \
    for (const auto& pcre : pcre_list) if (pcre.match(udsentry_path).hasMatch()) {udsentry_must_be_excluded = true; break;} \
    if (Q_UNLIKELY(udsentry_must_be_excluded)) continue; \
    for (const auto& path_to_exclude : path_to_exclude_list) if (udsentry_path.startsWith(path_to_exclude)) {udsentry_must_be_excluded = true; break;} \
    if (Q_UNLIKELY(udsentry_must_be_excluded)) continue

#define FILL_LIST(LIST, URL) \
    const auto& LIST##_job = KIO::listRecursive(URL, KIO::HideProgressInfo, include_hidden_files); \
    LIST##_job->setUiDelegate(nullptr); \
    LIST##_job->setUiDelegateExtension(nullptr); \
    QObject::connect(LIST##_job, &KIO::ListJob::entries, [&](void*, const KIO::UDSEntryList& partial_list){LIST.append(partial_list);}); \
    if (Q_UNLIKELY(LIST##_job->exec() == false)) { \
        results.error_message = LIST##_job->errorString(); \
        if (Q_UNLIKELY(results.error_message.isEmpty())) \
            results.error_message = LIST##_job->detailedErrorStrings().first(); \
        Q_ASSERT(results.error_message.isEmpty() == false); \
        return results; \
    } \
    if (Q_UNLIKELY(abort_preparation)) return results; \
    delete LIST##_job

#define INIT_QURL_FROM_OPERATION_DESTINATION_URL(NAME, RELATIVE_PATH) QUrl NAME(operation_urls.second); NAME.setPath(NAME.path() % RELATIVE_PATH)
#define INIT_QURL_FROM_OPERATION_SOURCE_URL(NAME, RELATIVE_PATH) QUrl NAME(operation_urls.first); NAME.setPath(NAME.path() % RELATIVE_PATH)

#define IS_DIR(UDSENTRY) UDSENTRY.isDir() == true && UDSENTRY.isLink() == false


Preparer::Results Preparer::parse(const OperationType& operation_type,
                                  const std::pair<QUrl,QUrl>& operation_urls,
                                  const bool& include_hidden_files,
                                  const QList<QRegularExpression>& pcre_list,
                                  QStringList path_to_exclude_list)
{
    Results results;

    // Retrieve udsentry lists
    KIO::UDSEntryList source_udsentry_list, destination_udsentry_list;
    {
        FILL_LIST(source_udsentry_list, operation_urls.first);
        if (Q_UNLIKELY(operation_type == OperationType::full_backup)) {
            INIT_QURL_FROM_OPERATION_DESTINATION_URL(destination_url, xi18nc("@item folder name", "Full backup of <filename>%1</filename> (%2)",
                                                                        operation_urls.first.fileName(), QDateTime::currentDateTime().toString(Qt::SystemLocaleLongDate)));
            results.task_deque.emplace_back(TaskType::Copy, operation_urls.first, std::move(destination_url));
            return results;
        }
        FILL_LIST(destination_udsentry_list, operation_urls.second);
    }

    // Parse lists
    const QLatin1String dot("."), dotdot("..");
    for (const auto& source_udsentry : source_udsentry_list) {
        bool found = false;
        const auto& udsentry_path = source_udsentry.stringValue(KIO::UDSEntry::UDS_NAME);
        APPLY_FILTERS;
        INIT_QURL_FROM_OPERATION_SOURCE_URL(source_udsentry_url, udsentry_path);
        INIT_QURL_FROM_OPERATION_DESTINATION_URL(destination_udsentry_url, udsentry_path);
        const auto& destination_udsentry_list_size = destination_udsentry_list.size();
        for (std::decay<decltype(destination_udsentry_list_size)>::type i = 0; i < destination_udsentry_list_size; ++i) {
            const auto& destination_udsentry = destination_udsentry_list.at(i);
            if (udsentry_path == destination_udsentry.stringValue(KIO::UDSEntry::UDS_NAME)) {
                const auto& source_udsentry_modification_time = source_udsentry.numberValue(KIO::UDSEntry::UDS_MODIFICATION_TIME);
                const auto& destination_udsentry_modification_time = destination_udsentry.numberValue(KIO::UDSEntry::UDS_MODIFICATION_TIME);
                if (source_udsentry_modification_time > destination_udsentry_modification_time) {
                    switch (operation_type) {
                    case OperationType::bilateral_differential_backup:
                    case OperationType::differential_backup:
                    case OperationType::synchronization:
                        if (IS_DIR(source_udsentry)) {
                            results.task_deque.emplace_back(TaskType::Delete, destination_udsentry_url, QUrl());
                            results.task_deque.emplace_back(TaskType::MakeDirectory, std::move(destination_udsentry_url), QUrl());
                        } else
                            results.task_deque.emplace_back(TaskType::Overwrite, std::move(source_udsentry_url), std::move(destination_udsentry_url));
                        break;
                    case OperationType::bilateral_incremental_backup:
                    case OperationType::incremental_backup:
                    {
                        INIT_QURL_FROM_OPERATION_DESTINATION_URL(backup, udsentry_path % QStringLiteral(" - ") % QDateTime::fromTime_t(destination_udsentry_modification_time).toString(Qt::SystemLocaleLongDate));
                        results.task_deque.emplace_back(TaskType::Rename, destination_udsentry_url, std::move(backup));
                        if (IS_DIR(source_udsentry))
                            results.task_deque.emplace_back(TaskType::MakeDirectory, std::move(destination_udsentry_url), QUrl());
                        else
                            results.task_deque.emplace_back(TaskType::Copy, std::move(source_udsentry_url), std::move(destination_udsentry_url));
                    } break;
                    case OperationType::full_backup:
                    case OperationType::none:
                        show_bug_and_exit(__FILE__, __LINE__);
                    }
                }
                if (source_udsentry_modification_time < destination_udsentry_modification_time) {
                    switch (operation_type) {
                    case OperationType::bilateral_differential_backup:
                        if (IS_DIR(source_udsentry)) {
                            results.task_deque.emplace_back(TaskType::Delete, source_udsentry_url, QUrl());
                            results.task_deque.emplace_back(TaskType::MakeDirectory, std::move(source_udsentry_url), QUrl());
                        } else
                            results.task_deque.emplace_back(TaskType::Overwrite, std::move(destination_udsentry_url), std::move(source_udsentry_url));
                        break;
                    case OperationType::bilateral_incremental_backup:
                    {
                        INIT_QURL_FROM_OPERATION_SOURCE_URL(backup, udsentry_path % QStringLiteral(" - ") % QDateTime::fromTime_t(source_udsentry_modification_time).toString(Qt::SystemLocaleLongDate));
                        results.task_deque.emplace_back(TaskType::Rename, source_udsentry_url, std::move(backup));
                        if (IS_DIR(source_udsentry))
                            results.task_deque.emplace_back(TaskType::MakeDirectory, std::move(source_udsentry_url), QUrl());
                        else
                            results.task_deque.emplace_back(TaskType::Copy, std::move(destination_udsentry_url), std::move(source_udsentry_url));
                    } break;
                    case OperationType::synchronization: // same as above
                        if (IS_DIR(source_udsentry)) {
                            results.task_deque.emplace_back(TaskType::Delete, destination_udsentry_url, QUrl());
                            results.task_deque.emplace_back(TaskType::MakeDirectory, std::move(destination_udsentry_url), QUrl());
                        } else
                            results.task_deque.emplace_back(TaskType::Overwrite, std::move(source_udsentry_url), std::move(destination_udsentry_url));
                        break;
                    case OperationType::differential_backup:
                    case OperationType::incremental_backup:
                        break; // skip
                    case OperationType::full_backup:
                    case OperationType::none:
                        show_bug_and_exit(__FILE__, __LINE__);
                    }
                }
                destination_udsentry_list.removeAt(i);
                found = true;
                break;
            }
        }
        if (found == false) {
            if (IS_DIR(source_udsentry))
                results.task_deque.emplace_back(TaskType::MakeDirectory, std::move(destination_udsentry_url), QUrl());
            else
                results.task_deque.emplace_back(TaskType::Copy, std::move(source_udsentry_url), std::move(destination_udsentry_url));
        }
    }
    if (Q_UNLIKELY(abort_preparation))
        return results;
    switch (operation_type) {
    case OperationType::bilateral_differential_backup:
    case OperationType::bilateral_incremental_backup:
        for (const auto& destination_udsentry : destination_udsentry_list) {
            const auto& udsentry_path = destination_udsentry.stringValue(KIO::UDSEntry::UDS_NAME);
            APPLY_FILTERS;
            INIT_QURL_FROM_OPERATION_DESTINATION_URL(destination_udsentry_url, udsentry_path);
            INIT_QURL_FROM_OPERATION_SOURCE_URL(source_udsentry_url, udsentry_path);
            if (IS_DIR(destination_udsentry))
                results.task_deque.emplace_back(TaskType::MakeDirectory, std::move(source_udsentry_url), QUrl());
            else
                results.task_deque.emplace_back(TaskType::Copy, std::move(destination_udsentry_url), std::move(source_udsentry_url));
        }
        break;
    case OperationType::synchronization:
        for (const auto& destination_udsentry : destination_udsentry_list) {
            const auto& udsentry_path = destination_udsentry.stringValue(KIO::UDSEntry::UDS_NAME);
            APPLY_FILTERS;
            INIT_QURL_FROM_OPERATION_DESTINATION_URL(destination_udsentry_url, udsentry_path);
            results.task_deque.emplace_front(TaskType::Delete, std::move(destination_udsentry_url), QUrl());
            if (IS_DIR(destination_udsentry))
                path_to_exclude_list.append(udsentry_path);
        }
        break;
    case OperationType::differential_backup:
    case OperationType::incremental_backup:
        break; // skip
    case OperationType::full_backup:
    case OperationType::none:
        show_bug_and_exit(__FILE__, __LINE__);
    }

    // Deliver results
    return results;
}


void Preparer::show_bug_and_exit(const char* const file_name, const int&& line_number)
{
    qFatal("%s, line %d: this point must never be reached!", file_name, line_number);
}
