/********************************************************************************
 * Copyright © 2014-2016 by Lorenzo Porta (Vindex17) <vindex17@outlook.it>      *
 *                                                                              *
 * This program is free software; you can redistribute it and/or                *
 * modify it under the terms of the GNU General Public License as               *
 * published by the Free Software Foundation; either version 2 of               *
 * the License or (at your option) version 3 or any later version               *
 * accepted by the membership of KDE e.V. (or its successor approved            *
 * by the membership of KDE e.V.), which shall act as a proxy                   *
 * defined in Section 14 of version 3 of the license.                           *
 *                                                                              *
 * This program is distributed in the hope that it will be useful,              *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                *
 * GNU General Public License for more details.                                 *
 *                                                                              *
 * You should have received a copy of the GNU General Public License            *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.        *
 *                                                                              *
 ********************************************************************************/


#include "urlrequester.h"
#include "urlrequester.moc"

#include <KIO/Global>
#include <KLocalizedString>
#include <QFileDialog>
#include <QHeaderView>

#define CHECK_RELATIONSHIP(A,B) if (Q_UNLIKELY(A##_url.isParentOf(B##_url))) if (Q_LIKELY(show_relationship_error_between(A##_string, B##_string) != KMessageBox::Continue)) return

#define SETUP_ITEM_FOR(X, COLUMN) \
    QTableWidgetItem* const X##_item = new QTableWidgetItem(QIcon::fromTheme(KIO::iconNameForUrl(X##_url)), X##_string); \
    X##_item->setFlags(Qt::ItemIsEnabled | Qt::ItemIsSelectable); \
    X##_item->setTextAlignment(Qt::AlignCenter); \
    setItem(new_row_n, COLUMN, X##_item);

#define SETUP_URL(NAME, WINDOW_TITLE) \
    const QUrl& NAME = QFileDialog::getExistingDirectoryUrl(this, WINDOW_TITLE, QUrl::fromLocalFile(QDir::homePath())).adjusted(QUrl::NormalizePathSegments); \
    if (Q_UNLIKELY(NAME.isEmpty())) return; \
    if (NAME.path().endsWith(slash) == false) const_cast<QUrl&>(NAME).setPath(NAME.path().append(slash))


UrlRequester::UrlRequester() : QTableWidget(0,2)
{
    horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
    verticalHeader()->setSectionResizeMode(QHeaderView::Fixed);
    setHorizontalHeaderLabels({i18nc("title:column", "Source directories"), i18nc("title:column", "Destination directories")});
    setSelectionBehavior(QAbstractItemView::SelectRows);
    setSelectionMode(QAbstractItemView::SingleSelection);
}


void UrlRequester::add()
{
    // Get and check new urls
    const auto& slash = QChar::fromLatin1('/');
    SETUP_URL(source_url, i18nc("@title:window", "Select the source directory"));
    SETUP_URL(destination_url, i18nc("@title:window", "Select the destination directory"));
    if (Q_UNLIKELY(source_url == destination_url)) {
        KMessageBox::sorry(this, i18nc("@info", "A directory can't be the source and the destination at the same time!"));
        return;
    }
    const auto& source_string = source_url.toDisplayString(URL_DISPLAY_FORMAT);
    const auto& destination_string = destination_url.toDisplayString(URL_DISPLAY_FORMAT);
    CHECK_RELATIONSHIP(source, destination);
    CHECK_RELATIONSHIP(destination, source);

    // Setup the new row
    const auto& new_row_n = rowCount();
    insertRow(new_row_n);
    SETUP_ITEM_FOR(source, 0);
    SETUP_ITEM_FOR(destination, 1);

    // Notify
    emit new_urls(std::make_pair(source_url, destination_url));
}


void UrlRequester::remove()
{
    const auto& i = currentRow();
    if (Q_UNLIKELY(i < 0))
        return;
    removeRow(i);
    emit deleted_row(i);
}


void UrlRequester::restore(const std::pair<QList<QUrl>,QList<QUrl>>& lists)
{
    const auto& lists_size = lists.first.size();
    setRowCount(lists_size);
    for (std::decay<decltype(lists_size)>::type i = 0; i < lists_size; ++i) {
        const auto& new_row_n = i; // for the macros
        const auto& source_url = lists.first.at(i);
        const auto& destination_url = lists.second.at(i);
        const auto& source_string = source_url.toDisplayString(URL_DISPLAY_FORMAT);
        const auto& destination_string = destination_url.toDisplayString(URL_DISPLAY_FORMAT);
        SETUP_ITEM_FOR(source, 0);
        SETUP_ITEM_FOR(destination, 1);
    }
}


KMessageBox::ButtonCode UrlRequester::show_relationship_error_between(const QString& parent_string, const QString& child_string)
{
    return KMessageBox::warningContinueCancel(this, xi18nc("@info", "<warning><filename>%1</filename> is parent of <filename>%2</filename>!</warning><nl/><nl/>"
                                                                    "<note>operations between related directories may open Pandora's box...</note><nl/>"
                                                                    "If you choose to continue, you take full responsibility for what you are doing.",
                                                           parent_string, child_string));
}
